import re
from typing import List

import html2text

from moodle_dl.types import Course
from moodle_dl.downloader.task import Task
from moodle_dl.utils import PathTools as PT


class TelegramFormater:
    @staticmethod
    def append_with_limit(new_line: str, one_msg_content: str, msg_list: List[str], limit: int = 4096):
        """Appends a new line to a message string,
        if the string is to long it ist appended to the message list.
        Returns the new message string.

        Args:
            new_line (str): A new line to append to a message string
            one_msg_content (str): The current message string
            msg_list ([str]): The list of finished messages
        Returns:
            str: The new message string
        """
        # Replace Telegram Entities
        new_line = re.sub("<(?!/b>|b>)", '&lt;', new_line)
        new_line = re.sub("(?<!</b)(?<!<b)>", '&gt;', new_line)
        if len(one_msg_content) + len(new_line) >= limit:
            msg_list.append(one_msg_content)
            if len(new_line) >= limit:
                new_line = new_line[: limit - 3] + '…'
            return new_line
        else:
            return one_msg_content + new_line

    @staticmethod
    def make_bold(string: str) -> str:
        """
        Makes a string bold in a telegram message
        """
        return '<b>' + string + '</b>'

    @classmethod
    def create_full_moodle_diff_messages(cls, changed_courses: List[Course]) -> List[str]:
        """
        Creates telegram messages with all changed files. This includes new,
        modified and deleted files. Files that have changed since the last message.

        @param changed_courses: A list of all courses with their modified files.
        @return: a list of messages
        """

        diff_count = 0
        for course in changed_courses:
            diff_count += len(course.files)

        result_list = []
        one_msg_content = f'{diff_count} new Changes in the Moodle courses!'

        for course in changed_courses:
            new_line = '\r\n\r\n\r\n👉 ' + cls.make_bold(course.fullname) + '\r\n'
            one_msg_content = cls.append_with_limit(new_line, one_msg_content, result_list)

            for file in course.files:
                saved_to_path = file.saved_to
                if file.new_file is not None:
                    saved_to_path = file.new_file.saved_to
                if file.modified:
                    new_line = '\r\n🔄 Modified:\r\n ' + saved_to_path
                elif file.moved:
                    new_line = '\r\n🔀 Moved:\r\n ' + saved_to_path
                elif file.deleted:
                    new_line = '\r\n❌ Deleted:\r\n ' + saved_to_path
                else:
                    new_line = '\r\n✅ Added:\r\n ' + saved_to_path

                one_msg_content = cls.append_with_limit(new_line, one_msg_content, result_list)

                if (not file.moved and not file.deleted) and (
                    file.content_type == 'description'
                    or (file.content_type == 'html' and file.module_modname == 'page')
                ):
                    text_lines = []
                    try:
                        with open(saved_to_path, 'r', encoding='utf-8') as content_file:
                            content = content_file.read()

                            if file.content_type == 'html':
                                h2t_handler = html2text.HTML2Text()
                                content = h2t_handler.handle(content).strip()
                                if content == '':
                                    continue

                            text_lines = content.splitlines()
                    except OSError:
                        text_lines = []

                    if len(text_lines) > 0:
                        one_msg_content = cls.append_with_limit('\r\n👇\r\n', one_msg_content, result_list)
                        for new_line in text_lines:
                            new_line = new_line + '\r\n'
                            one_msg_content = cls.append_with_limit(new_line, one_msg_content, result_list)
                        one_msg_content = cls.append_with_limit('👆\r\n', one_msg_content, result_list)

        result_list.append(one_msg_content)
        return result_list

    @classmethod
    def create_full_error_messages(cls, details) -> List[str]:
        """
        Creates error messages
        """
        result_list = []

        one_msg_content = '🛑 The following error occurred during execution:\r\n'
        for new_line in details.splitlines():
            new_line = new_line + '\r\n'
            one_msg_content = cls.append_with_limit(new_line, one_msg_content, result_list)

        result_list.append(one_msg_content)
        return result_list

    @classmethod
    def create_full_failed_downloads_messages(cls, failed_downloads: List[Task]) -> List[str]:
        """
        Creates messages with all failed downloads
        """

        result_list = []
        if len(failed_downloads) == 0:
            return result_list

        one_msg_content = (
            '⁉️ Error while trying to download files, look at the log for more details.'
            + '\r\nList of failed downloads:\r\n\r\n'
        )
        for task in failed_downloads:
            new_line = (
                f'⚠️ {PT.to_valid_name(task.file.content_filename, is_file=True)}:'
                + f'\r\n{task.status.get_error_text()}\r\n\r\n'
            )
            if task.file.content_filename != task.file.content_fileurl:
                new_line = (
                    f'⚠️ {PT.to_valid_name(task.file.content_filename, is_file=True)} ({task.file.content_fileurl}):'
                    + f'\r\n{task.status.get_error_text()}\r\n\r\n'
                )

            one_msg_content = cls.append_with_limit(new_line, one_msg_content, result_list)

        result_list.append(one_msg_content)
        return result_list



import logging
import re

from typing import Dict, List

from moodle_dl.config import ConfigHelper
from moodle_dl.moodle.mods import MoodleMod
from moodle_dl.moodle.moodle_constants import moodle_html_footer, moodle_html_header
from moodle_dl.moodle.request_helper import RequestRejectedError
from moodle_dl.types import Course, File
from moodle_dl.utils import PathTools as PT


class LessonMod(MoodleMod):
    MOD_NAME = 'lesson'
    MOD_PLURAL_NAME = 'lessons'
    MOD_MIN_VERSION = 2017051500  # 3.3

    @classmethod
    def download_condition(cls, config: ConfigHelper, file: File) -> bool:
        return config.get_download_lessons() or (not (file.module_modname.endswith(cls.MOD_NAME) and file.deleted))

    async def real_fetch_mod_entries(self, courses: List[Course]) -> Dict[int, Dict[int, Dict]]:
        lessons = (
            await self.client.async_post(
                'mod_lesson_get_lessons_by_courses', self.get_data_for_mod_entries_endpoint(courses)
            )
        ).get('lessons', [])

        result = {}
        for lesson in lessons:
            course_id = lesson.get('course', 0)
            lesson_files = lesson.get('introfiles', [])
            lesson_files += lesson.get('mediafiles', [])
            self.set_files_types_if_empty(lesson_files, 'lesson_introfile')

            lesson_intro = lesson.get('intro', '')
            if lesson_intro != '':
                lesson_files.append(
                    {
                        'filename': 'Lesson intro',
                        'filepath': '/',
                        'description': lesson_intro,
                        'type': 'description',
                    }
                )

            self.add_module(
                result,
                course_id,
                lesson.get('coursemodule', 0),
                {
                    'id': lesson.get('id', 0),
                    'name': lesson.get('name', 'unnamed lesson'),
                    'files': lesson_files,
                },
            )

        await self.add_lessons_files(result)
        return result

    async def add_lessons_files(self, lessons: Dict[int, Dict[int, Dict]]):
        """
        Fetches for the lessons list the lessons files
        @param lessons: Dictionary of all lessons, indexed by courses, then module id
        """
        if not self.config.get_download_lessons():
            return

        if self.version < 2017051500:  # 3.3
            return

        await self.run_async_load_function_on_mod_entries(lessons, self.load_lesson_files)

    async def load_lesson_files(self, lesson: Dict):
        # load only the last attempt
        # TODO: We could load all attempts, if needed
        lesson_id = lesson.get('id', 0)
        data = {'lessonid': lesson_id, 'userid': self.user_id, 'lessonattempt': 0}
        try:
            user_attempt = await self.client.async_post('mod_lesson_get_user_attempt', data)
            user_attempt['lesson_name'] = lesson.get('name', '')
        except RequestRejectedError:
            logging.debug("No access rights for lesson %d", lesson_id)
            return
        lesson['files'] += await self._get_files_of_attempt(user_attempt)

    async def load_attempt_page(self, attempt_page: Dict) -> List[Dict]:
        "Load files and content of an answer page in an attempt"
        page_id = attempt_page.get('page', {}).get('id', 0)
        data = {'lessonid': attempt_page.get('page', {}).get('lessonid', 0), 'pageid': page_id, 'returncontents': 1}
        try:
            page_result = await self.client.async_post('mod_lesson_get_page_data', data)
        except RequestRejectedError:
            logging.debug("No access rights for lesson attempt page %d", page_id)
            return []

        page_files = page_result.get('contentfiles', [])
        self.set_files_types_if_empty(page_files, 'lesson_file')

        page_files.append(
            {
                '_is_page_content': True,
                'content': page_result.get('pagecontent', '').split('<script>')[0],
            }
        )
        return page_files

    async def _get_files_of_attempt(self, attempt: Dict) -> List[Dict]:
        result = []

        # Calculate last modified
        completed = attempt.get('userstats', {}).get('completed', 0)
        for answerpage in attempt.get('answerpages', []):
            answerpage_page = answerpage.get('page', {})
            last_modified = answerpage_page.get('timemodified', 0)
            if last_modified == 0:
                last_modified = answerpage_page.get('timecreated', 0)
            if completed < last_modified:
                completed = last_modified

        # Create grade file
        grade = attempt.get('userstats', {}).get('gradeinfo', {}).get('earned', None)
        grade_total = attempt.get('userstats', {}).get('gradeinfo', {}).get('total', None)
        if grade is not None and grade_total is not None:
            result.append(
                {
                    'filename': 'grade',
                    'filepath': '/',
                    'timemodified': completed,
                    'description': str(grade) + ' / ' + str(grade_total),
                    'type': 'description',
                }
            )

        attempt_pages_and_files = await self.run_async_collect_function_on_list(
            attempt.get('answerpages', []),
            self.load_attempt_page,
            'attempt page',
            {'collect_id': 'page.id', 'collect_name': 'page.lessonid'},
        )

        # build lesson HTML and add unique files
        lesson_html = moodle_html_header
        lesson_is_empty = True
        for page_or_file in attempt_pages_and_files:
            if page_or_file.get('_is_page_content', False):
                page_content = page_or_file.get('content', '')
                if page_content != '':
                    lesson_is_empty = False
                lesson_html += page_content + '\n'
            else:
                new_page_file = True
                clean_file_url = re.sub(r"\/page_contents\/\d+\/", "/", page_or_file.get('fileurl', ''))
                page_or_file['_clean_file_url'] = clean_file_url
                for attempt_file in result:
                    if attempt_file.get('_clean_file_url', '') == clean_file_url:
                        # sometimes the teacher adds the same file for multiple answer pages with a
                        # different timestamp
                        if (attempt_file.get('filesize', 0) == page_or_file.get('filesize', 0)) and (
                            attempt_file.get('filename', '') == page_or_file.get('filename', '')
                        ):
                            new_page_file = False
                            break
                if new_page_file:
                    result.append(page_or_file)

        # The generation code for the original review page is here:
        # https://github.com/moodle/moodle/blob/511a87f5fc357f18a4c53911f6e6c7f7b526246e/mod/lesson/report.php#L278-L366
        if not lesson_is_empty:
            lesson_html += moodle_html_footer
            result.append(
                {
                    'filename': PT.to_valid_name(attempt['lesson_name'], is_file=False),
                    'filepath': '/',
                    'timemodified': completed,
                    'html': lesson_html,
                    'filter_urls_during_search_containing': ['/mod_lesson/page_contents/'],
                    'type': 'html',
                    'no_search_for_urls': True,
                }
            )

        return result



import asyncio
import json
import logging
import os
import urllib

from time import sleep
from typing import Dict

import aiohttp
import requests

from requests.exceptions import RequestException

from moodle_dl.types import MoodleURL, MoodleDlOpts
from moodle_dl.utils import SslHelper, PathTools as PT, MoodleDLCookieJar


class RequestHelper:
    """
    Functions for sending out requests to the Moodle System.
    """

    RQ_HEADER = {
        'User-Agent': (
            'Mozilla/5.0 (Linux; Android 7.1.1; Moto G Play Build/NPIS26.48-43-2; wv) AppleWebKit/537.36'
            + ' (KHTML, like Gecko) Version/4.0 Chrome/71.0.3578.99 Mobile Safari/537.36 MoodleMobile'
        ),
        'Content-Type': 'application/x-www-form-urlencoded',
    }
    MAX_RETRIES = 5

    def __init__(self, opts: MoodleDlOpts, moodle_url: MoodleURL, token: str = ''):
        self.token = token
        self.moodle_url = moodle_url
        self.opts = opts

        self.url_base = moodle_url.url_base

        # Semaphore for async requests
        # Keep in mind Semaphore needs to be initialized in the same async loop as it is used
        self.semaphore = asyncio.Semaphore(opts.max_parallel_api_calls)

        self.log_responses_to = None
        if opts.log_responses:
            self.log_responses_to = PT.make_path(opts.path, 'responses.log')
            with open(self.log_responses_to, 'w', encoding='utf-8') as response_log_file:
                response_log_file.write('JSON Log:\n\n')

    def post_URL(self, url: str, data: Dict[str, str] = None, cookie_jar_path: str = None):
        """
        Sends a POST request to a specific URL, including saving of cookies in cookie jar.
        @param url: The url to which the request is sent. (the moodle base url is not added to the given URL)
        @param data: The optional data is added to the POST body.
        @param cookie_jar_path: Path to the cookies file.
        @return: The resulting response object and the session object.
        """

        data_urlencoded = ""
        if data is not None:
            data_urlencoded = RequestHelper.recursive_urlencode(data)

        session = SslHelper.custom_requests_session(self.opts.skip_cert_verify, self.opts.allow_insecure_ssl)
        if cookie_jar_path is not None:
            session.cookies = MoodleDLCookieJar(cookie_jar_path)

            if os.path.exists(cookie_jar_path):
                session.cookies.load(ignore_discard=True, ignore_expires=True)
        try:
            response = session.post(url, data=data_urlencoded, headers=self.RQ_HEADER, timeout=60)
        except RequestException as error:
            raise ConnectionError(f"Connection error: {str(error)}") from None

        if cookie_jar_path is not None:
            for cookie in session.cookies:
                cookie.expires = 2147483647

            session.cookies.save(ignore_discard=True, ignore_expires=True)

        return response, session

    def get_URL(self, url: str, cookie_jar_path: str = None):
        """
        Sends a GET request to a specific URL of the Moodle system, including additional cookies
        (cookies are updated after the request)
        @param url: The url to which the request is sent. (the moodle base url is not added to the given URL)
        @param cookie_jar_path: The optional cookies to add to the request
        @return: The resulting Response object.
        """

        session = SslHelper.custom_requests_session(self.opts.skip_cert_verify, self.opts.allow_insecure_ssl)
        if cookie_jar_path is not None:
            session.cookies = MoodleDLCookieJar(cookie_jar_path)

            if os.path.exists(cookie_jar_path):
                session.cookies.load(ignore_discard=True, ignore_expires=True)
            session.cookies = session.cookies
        try:
            response = session.get(url, headers=self.RQ_HEADER, timeout=60)
        except RequestException as error:
            raise ConnectionError(f"Connection error: {str(error)}") from None

        if cookie_jar_path is not None:
            session.cookies.save(ignore_discard=True, ignore_expires=True)

        return response, session

    async def async_post(self, function: str, data: Dict[str, str] = None, timeout: int = 60) -> Dict:
        """
        Sends async a POST request to the REST endpoint of the Moodle system
        @param function: The Web service function to be called.
        @param data: The optional data is added to the POST body.
        @return: The JSON response returned by the Moodle system, already checked for errors..
        """

        if self.token is None:
            raise ValueError('The required token is not set!')

        data_urlencoded = self._get_POST_DATA(function, self.token, data)
        url = self._get_REST_POST_URL(self.url_base, function)
        ssl_context = SslHelper.get_ssl_context(self.opts.skip_cert_verify, self.opts.allow_insecure_ssl)

        error_ctr = 0
        async with self.semaphore, aiohttp.ClientSession() as session:
            while error_ctr < self.MAX_RETRIES:
                try:
                    async with session.post(
                        url,
                        data=data_urlencoded,
                        headers=self.RQ_HEADER,
                        timeout=timeout,
                        ssl=ssl_context,
                        raise_for_status=True,
                    ) as resp:
                        resp_json = await resp.json()
                        resp_url = resp.url
                    break
                except (aiohttp.client_exceptions.ClientError, OSError, ValueError) as req_err:
                    if (isinstance(req_err, aiohttp.client_exceptions.ClientResponseError)) and (
                        req_err.status not in [408, 409, 429]  # pylint: disable=no-member
                    ):
                        # 408 (timeout) or 409 (conflict) and 429 (too many requests)
                        raise ConnectionError(f"Connection error: {req_err}") from None
                    if isinstance(req_err, aiohttp.client_exceptions.ContentTypeError):
                        raise RequestRejectedError(
                            'The Moodle Mobile API does not appear to be available at this time.'
                        ) from None

                    error_ctr += 1
                    if error_ctr < self.MAX_RETRIES:
                        logging.debug("The %sth connection error occurred, retrying. %s", error_ctr, req_err)
                        asyncio.sleep(1)
                        continue
                    raise ConnectionError(f"Connection error: {req_err}") from None

        self.check_json_for_moodle_error(resp_json)
        self.log_response(function, data, resp_url, resp_json)

        return resp_json

    def post(self, function: str, data: Dict[str, str] = None, timeout: int = 60) -> Dict:
        """
        Sends a POST request to the REST endpoint of the Moodle system
        @param function: The Web service function to be called.
        @param data: The optional data is added to the POST body.
        @return: The JSON response returned by the Moodle system, already checked for errors..
        """

        if self.token is None:
            raise ValueError('The required Token is not set!')

        data_urlencoded = self._get_POST_DATA(function, self.token, data)
        url = self._get_REST_POST_URL(self.url_base, function)

        session = SslHelper.custom_requests_session(self.opts.skip_cert_verify, self.opts.allow_insecure_ssl)
        error_ctr = 0
        while error_ctr <= self.MAX_RETRIES:
            try:
                response = session.post(url, data=data_urlencoded, headers=self.RQ_HEADER, timeout=timeout)
                break
            except (requests.ConnectionError, requests.Timeout) as req_err:
                # We treat requests.ConnectionErrors here specially, since they normally mean,
                # that something went wrong, which could be fixed by a restart.
                error_ctr += 1
                if error_ctr < self.MAX_RETRIES:
                    logging.debug("The %sth Connection Error occurred, retrying. %s", error_ctr, str(req_err))
                    sleep(1)
                    continue
                raise ConnectionError(f"Connection error: {req_err}") from None
            except RequestException as req_err:
                raise ConnectionError(f"Connection error: {req_err}") from None

        json_result = self._initial_parse(response)
        self.log_response(function, data, response.url, json_result)

        return json_result

    def log_response(self, function: str, data: Dict[str, str], url: str, json_result: Dict):
        if self.opts.log_responses and function not in ['tool_mobile_get_autologin_key']:
            with open(self.log_responses_to, 'a', encoding='utf-8') as response_log_file:
                response_log_file.write(f'URL: {url}\n')
                response_log_file.write(f'Function: {function}\n\n')
                response_log_file.write(f'Data: {data}\n\n')
                response_log_file.write(json.dumps(json_result, indent=4, ensure_ascii=False))
                response_log_file.write('\n\n\n')

    @staticmethod
    def _get_REST_POST_URL(url_base: str, function: str) -> str:
        """
        Generates an URL for a REST-POST request
        @params: The necessary parameters for a REST URL
        @return: A formatted URL
        """
        url = f'{url_base}webservice/rest/server.php?moodlewsrestformat=json&wsfunction={function}'

        return url

    @staticmethod
    def _get_POST_DATA(function: str, token: str, data_obj: str) -> str:
        """
        Generates the data for a REST-POST request
        @params: The necessary parameters for a REST URL
        @return: A URL-encoded data string
        """
        data = {'moodlewssettingfilter': 'true', 'moodlewssettingfileurl': 'true'}

        if data_obj is not None:
            data.update(data_obj)

        data.update({'wsfunction': function, 'wstoken': token})

        return RequestHelper.recursive_urlencode(data)

    def get_login(self, data: Dict[str, str]) -> object:
        """
        Sends a POST request to the login endpoint of the Moodle system to
        obtain a token in JSON format.
        @param data: The data is inserted into the Post-Body as arguments. This
        should contain the login data.
        @return: The JSON response returned by the Moodle System, already
        checked for errors.
        """
        session = SslHelper.custom_requests_session(self.opts.skip_cert_verify, self.opts.allow_insecure_ssl)
        try:
            response = session.post(
                f'{self.url_base}login/token.php',
                data=urllib.parse.urlencode(data),
                headers=self.RQ_HEADER,
                timeout=60,
            )
        except RequestException as error:
            raise ConnectionError(f"Connection error: {str(error)}") from None

        return self._initial_parse(response)

    @staticmethod
    def _check_response_code(response):
        # Normally Moodle answer with response 200
        if response.status_code != 200:
            raise RequestRejectedError(
                'An Unexpected Error happened on side of the Moodle System!'
                + f' Status-Code: {str(response.status_code)}'
                + f'\nHeader: {response.headers}'
                + f'\nResponse: {response.text}'
            )

    def _initial_parse(self, response) -> object:
        """
        The first time parsing the result of a REST request.
        It is checked for known errors.
        @param response: The JSON response of the Moodle system
        @return: The parsed JSON object
        """

        self._check_response_code(response)

        # Try to parse the JSON
        try:
            resp_json = response.json()
        except ValueError:
            raise RequestRejectedError('The Moodle Mobile API does not appear to be available at this time.') from None
        except Exception as error:
            raise RequestRejectedError(
                'An Unexpected Error occurred while trying'
                + ' to parse the json response! Moodle'
                + f' response: {response.text}.\nError: {error}'
            ) from None

        self.check_json_for_moodle_error(resp_json)
        return resp_json

    def check_json_for_moodle_error(self, resp_json: Dict):
        # Check for known errors
        if 'error' in resp_json:
            raise RequestRejectedError(
                'The Moodle System rejected the Request.'
                + f" Details: {resp_json.get('error', '')} (Errorcode: {resp_json.get('errorcode', '')},"
                + f" Stacktrace: {resp_json.get('stacktrace', '')}, Debuginfo: {resp_json.get('debuginfo', '')},"
                + f" Reproductionlink: {resp_json.get('reproductionlink', '')})"
            )

        if 'exception' in resp_json:
            errorcode = resp_json.get('errorcode', '')

            if errorcode == 'invalidtoken':
                raise RequestRejectedError(
                    'Your Moodle token has expired.'
                    + ' To create a new one run "moodle-dl -nt -u USERNAME -pw PASSWORD" or "moodle-dl -nt -sso"'
                )

            raise RequestRejectedError(
                'The Moodle System rejected the Request.'
                + f" Details: {resp_json.get('exception', '')} (Errorcode: {errorcode},"
                + f" Message: {resp_json.get('message', '')})"
            )

    @staticmethod
    def recursive_urlencode(data):
        """URL-encode a multidimensional dictionary.
        @param data: the data to be encoded
        @return: the url encoded data
        """

        def recursion(data, base=None):
            if base is None:
                base = []
            pairs = []

            for key, value in data.items():
                new_base = base + [key]
                if hasattr(value, 'values'):
                    pairs += recursion(value, new_base)
                else:
                    new_pair = None
                    if len(new_base) > 1:
                        first = urllib.parse.quote(new_base.pop(0))
                        rest = map(urllib.parse.quote, new_base)
                        new_pair = f"{first}[{']['.join(rest)}]={urllib.parse.quote(str(value))}"
                    else:
                        new_pair = f'{urllib.parse.quote(str(key))}={urllib.parse.quote(str(value))}'
                    pairs.append(new_pair)
            return pairs

        return '&'.join(recursion(data))


class RequestRejectedError(Exception):
    """An Exception which gets thrown if the Moodle-System answered with an
    Error to our Request"""

    pass



import logging
import traceback

from typing import List

from moodle_dl.downloader.task import Task
from moodle_dl.notifications.mail.mail_shooter import MailShooter
from moodle_dl.notifications.notification_service import NotificationService
from moodle_dl.types import Course
from moodle_dl.notifications.mail.mail_formater import (
    create_full_error_mail,
    create_full_failed_downloads_mail,
    create_full_moodle_diff_mail,
)


class MailService(NotificationService):
    def _is_configured(self) -> bool:
        # Checks if the sending of emails has been configured.
        try:
            self.config.get_property('mail')
            return True
        except ValueError:
            logging.debug('Mail-Notifications not configured, skipping.')
            return False

    def _send_mail(self, subject, mail_content: (str, {str: str})):
        """
        Sends an email
        """
        if not self._is_configured():
            return

        mail_cfg = self.config.get_property('mail')

        try:
            logging.info('Sending Notification via Mail...')

            mail_shooter = MailShooter(
                mail_cfg['sender'],
                mail_cfg['server_host'],
                int(mail_cfg['server_port']),
                mail_cfg['username'],
                mail_cfg['password'],
            )
            mail_shooter.send(mail_cfg['target'], subject, mail_content[0], mail_content[1])
        except BaseException as e:
            logging.error('While sending notification:\n%s', traceback.format_exc(), extra={'exception': e})
            raise e  # to be properly notified via Sentry

    def notify_about_changes_in_moodle(self, changes: List[Course]) -> None:
        """
        Sends out a notification email about the downloaded changes.
        @param changes: A list of changed courses with changed files.
        """
        if not self._is_configured():
            return

        mail_content = create_full_moodle_diff_mail(changes)

        diff_count = 0
        for course in changes:
            diff_count += len(course.files)

        self._send_mail(f'{diff_count} new Changes in the Moodle courses!', mail_content)

    def notify_about_error(self, error_description: str):
        """
        Sends out an error mail if configured to do so.
        @param error_description: The error object.
        """
        if not self._is_configured():
            return

        mail_cfg = self.config.get_property('mail')

        if not mail_cfg.get('send_error_msg', True):
            return

        mail_content = create_full_error_mail(error_description)
        self._send_mail('Error!', mail_content)

    def notify_about_failed_downloads(self, failed_downloads: List[Task]):
        """
        Sends out an mail with all failed download if configured to send out error messages.
        @param failed_downloads: A list of failed Tasks.
        """
        if not self._is_configured():
            return

        mail_cfg = self.config.get_property('mail')

        if not mail_cfg.get('send_error_msg', True):
            return

        mail_content = create_full_failed_downloads_mail(failed_downloads)
        self._send_mail('Faild to download files!', mail_content)



from typing import List

from moodle_dl.downloader.task import Task
from moodle_dl.notifications.notification_service import NotificationService
from moodle_dl.types import Course
from moodle_dl.utils import Log, PathTools as PT


class ConsoleService(NotificationService):
    def notify_about_changes_in_moodle(self, changes: List[Course]) -> None:
        """
        Creates a terminal output about the downloaded changes.
        @param changes: A list of changed courses with changed files.
        """
        print('\n')

        diff_count = 0
        for course in changes:
            diff_count += len(course.files)

        if diff_count > 0:
            Log.success(f'{diff_count} changes found for the configured Moodle-Account.')

        for course in changes:
            if len(course.files) == 0:
                continue

            Log.blue(course.fullname)

            for file in course.files:
                saved_to_path = file.saved_to
                if file.new_file is not None:
                    saved_to_path = file.new_file.saved_to
                if file.modified:
                    Log.yellow('≠\t' + saved_to_path)
                elif file.moved:
                    if file.new_file is not None:
                        print(Log.cyan_str('<->\t' + file.saved_to) + Log.green_str(' ==> ' + saved_to_path))

                    else:
                        print(Log.cyan_str('<->\t' + saved_to_path))

                elif file.deleted:
                    print(Log.magenta_str('-\t' + saved_to_path))

                else:
                    print(Log.green_str('+\t' + saved_to_path))
            print('\n')

    def notify_about_error(self, error_description: str):
        Log.error(f'The following error occurred during execution:\n{error_description}')

    def notify_about_failed_downloads(self, failed_downloads: List[Task]):
        if len(failed_downloads) > 0:
            print('')
            Log.warning(
                'Error while trying to download files, look at the log for more details. List of failed downloads:'
            )
            print('')

        for task in failed_downloads:
            Log.cyan(PT.to_valid_name(task.file.content_filename, is_file=True))
            Log.error(f'\t{task.status.get_error_text()}')

        print('')



import logging
import os
import platform

from pathlib import Path
from typing import List

from moodle_dl.config import ConfigHelper
from moodle_dl.database import StateRecorder
from moodle_dl.downloader.task import Task
from moodle_dl.types import Course, MoodleDlOpts
from moodle_dl.utils import PathTools as PT


class FakeDownloadService:
    """
    FakeDownloadService fakes a DownloadService.
    This way a local database of Moodle's current files can be created without actually downloading the files.
    """

    def __init__(self, courses: List[Course], config: ConfigHelper, opts: MoodleDlOpts, database: StateRecorder):
        self.courses = courses
        self.opts = opts
        self.config = config
        self.database = database

    def get_failed_tasks(self):
        """
        Return a list of failed downloads.
        No download can fail, so this is only a dummy function.
        """
        return []

    def run(self):
        # delete files, that should be deleted
        self.database.batch_delete_files(self.courses)

        # save files, that should be saved
        for course in self.courses:
            for file in course.files:
                if file.deleted is False:
                    save_destination = Task.gen_path(self.opts.path, course, file)
                    filename = PT.to_valid_name(file.content_filename, is_file=True)

                    if file.content_type == 'description':
                        file.saved_to = str(Path(save_destination) / (filename + '.md'))

                    elif file.content_type == 'html':
                        file.saved_to = str(Path(save_destination) / (filename + '.html'))

                    elif file.module_modname.startswith('url'):
                        file.saved_to = str(Path(save_destination) / (filename + '.desktop'))
                        if os.name == 'nt' or platform.system() == "Darwin":
                            file.saved_to = str(Path(save_destination) / (filename + '.URL'))
                    else:
                        file.saved_to = str(Path(save_destination) / filename)

                    self.database.save_file(file, course.id, course.fullname)

        logging.info('All files stored in the Database!')
